@@ -1,4 +1,6 @@
 """
+parsr
+=====
 parsr is a library for building parsers based on `parsing expression grammars or
 PEGs <http://bford.info/pub/lang/peg.pdf>`_.
 
@@ -47,6 +49,7 @@ Here's an example that evaluates arithmetic expressions.
 from __future__ import print_function
 import functools
 import logging
+import re
 import string
 from bisect import bisect_left
 from six import StringIO, with_metaclass
@@ -120,16 +123,16 @@ def _debug_hook(func):
         if self._debug:
             line = ctx.line(pos) + 1
             col = ctx.col(pos) + 1
-            log.debug("Trying {} at line {} col {}".format(self, line, col))
+            log.debug("Trying {0} at line {1} col {2}".format(self, line, col))
         try:
             res = func(self, pos, data, ctx)
             if self._debug:
-                log.debug("Result: {}".format(res[1]))
+                log.debug("Result: {0}".format(res[1]))
             return res
         except:
             if self._debug:
                 ps = "-> ".join([str(p) for p in ctx.parser_stack])
-                log.debug("Failed: {}".format(ps))
+                log.debug("Failed: {0}".format(ps))
             raise
         finally:
             ctx.parser_stack.pop()
@@ -143,11 +146,12 @@ class Context(object):
     for grammars like xml or apache configuration, the active parser stack for
     error reporting, and accumulated errors for the farthest position reached.
     """
-    def __init__(self, lines, src=None):
+    def __init__(self, lines, orig, src=None):
         self.pos = -1
         self.indents = []
         self.tags = []
         self.src = src
+        self.orig = orig
         self.lines = [i for i, x in enumerate(lines) if x == "\n"]
         self.parser_stack = []
         self.errors = []
@@ -220,12 +224,12 @@ class Parser(with_metaclass(_ParserMeta, Node)):
         """
         return Lift(self._accumulate) * Opt(self) * Many(sep >> self)
 
-    def until(self, pred):
+    def until(self, pred, upper=None):
         """
         Return an :py:class:`Until` parser that matches zero or more instances
         of the current parser until the pred parser succeeds.
         """
-        return Until(self, pred)
+        return Until(self, pred, upper=upper)
 
     def map(self, func):
         """
@@ -309,12 +313,12 @@ class Parser(with_metaclass(_ParserMeta, Node)):
         parsers have particular needs not covered by the default
         implementation that provides significant indent and tag stacks.
         """
-        data = list(data)
-        data.append(None)  # add a terminal so we don't overrun
-        ctx = Ctx(data, src=src)
+        chars = list(data)
+        chars.append(None)  # add a terminal so we don't overrun
+        ctx = Ctx(chars, data, src=src)
 
         try:
-            _, ret = self.process(0, data, ctx)
+            _, ret = self.process(0, chars, ctx)
             return ret
         except Exception:
             pass
@@ -323,13 +327,13 @@ class Parser(with_metaclass(_ParserMeta, Node)):
 
         lineno = ctx.line(ctx.pos) + 1
         colno = ctx.col(ctx.pos) + 1
-        msg = "At line {} column {}:"
+        msg = "At line {0} column {1}:"
         print(msg.format(lineno, colno, ctx.lines), file=err)
         for parsers, msg in ctx.errors:
             names = " -> ".join([p.name for p in parsers if p.name])
             v = data[ctx.pos] or "EOF"
             print(names, file=err)
-            print("    {} Got {!r}.".format(msg, v), file=err)
+            print("    {0} Got {1!r}.".format(msg, v), file=err)
         err.seek(0)
         raise Exception(err.read())
 
@@ -361,12 +365,12 @@ class Char(Parser):
     def __init__(self, char):
         super(Char, self).__init__()
         self.char = char
-        self.name = "Char({!r})".format(self.char)
+        self.name = "Char({0!r})".format(self.char)
 
     def process(self, pos, data, ctx):
         if data[pos] == self.char:
             return (pos + 1, self.char)
-        msg = "Expected {!r}.".format(self.char)
+        msg = "Expected {0!r}.".format(self.char)
         ctx.set(pos, msg)
         raise Exception(msg)
 
@@ -398,13 +402,13 @@ class InSet(Parser):
         c = data[pos]
         if c in self.values:
             return (pos + 1, c)
-        msg = "Expected {}.".format(self)
+        msg = "Expected {0}.".format(self)
         ctx.set(pos, msg)
         raise Exception(msg)
 
     def __repr__(self):
         if self.name is None:
-            return "InSet({!r})".format(sorted(self.values))
+            return "InSet({0!r})".format(sorted(self.values))
         return super(InSet, self).__repr__()
 
 
@@ -444,12 +448,65 @@ class String(Parser):
                 break
             p = data[pos]
         if len(results) < self.min_length:
-            msg = "Expected {} of {}.".format(self.min_length, sorted(self.chars))
+            msg = "Expected {0} of {1}.".format(self.min_length, sorted(self.chars))
             ctx.set(old, msg)
             raise Exception(msg)
         return pos, "".join(results)
 
 
+class StringUntil(Parser):
+    """
+    StringUntil matches any number of characters until a predicate is seen.
+
+        .. code-block:: python
+
+            su  = StringUntil(Char("="))  # parses any number of characters until 'a'
+            val = su("ab=")               # produces "ab" from the data.
+            val = su("ab")                # raises an exception
+
+    """
+    def __init__(self, term, lower=None, upper=None):
+        super(StringUntil, self).__init__()
+        self.add_child(AnyChar.until(term, upper=upper).map(lambda x: "".join(x)) & term)
+        self.lower = lower
+
+    def process(self, pos, data, ctx):
+        (newpos, res) = self.children[0].process(pos, data, ctx)
+        if self.lower is not None and len(res) < self.lower:
+            ctx.set(pos, "Expected at least {0} characters.".format(self.lower))
+            raise Exception()
+        return newpos, res
+
+
+class Regex(Parser):
+    """
+    Match characters against a regular expression.
+
+        .. code-block:: python
+
+            identifier = Regex("[a-zA-Z]([a-zA-Z0-9])*")
+            identifier("abcd1") # returns "abcd1"
+            identifier("1bcd1") # raises an exception
+
+    """
+    def __init__(self, pattern, flags=0, return_match=False):
+        super(Regex, self).__init__()
+        self.pattern = pattern
+        self.flags = flags
+        self.regex = re.compile(pattern, flags)
+        self.return_match = return_match
+
+    def process(self, pos, data, ctx):
+        try:
+            m = self.regex.match(ctx.orig[pos:])
+            start, end = m.span()
+            res = m if self.return_match else ctx.orig[pos:pos + end]
+            return pos + end, res
+        except:
+            ctx.set(pos, "Expected pattern {0!r} (flags={1}).".format(self.pattern, self.flags))
+            raise Exception()
+
+
 class Literal(Parser):
     """
     Match a literal string. The ``value`` keyword lets you return a python
@@ -491,7 +548,7 @@ class Literal(Parser):
         self.chars = chars if not ignore_case else chars.lower()
         self.value = value
         self.ignore_case = ignore_case
-        self.name = "Literal{!r}".format(self.chars)
+        self.name = "Literal{0!r}".format(self.chars)
 
     def process(self, pos, data, ctx):
         old = pos
@@ -500,7 +557,7 @@ class Literal(Parser):
                 if data[pos] == c:
                     pos += 1
                 else:
-                    msg = "Expected {!r}.".format(self.chars)
+                    msg = "Expected {0!r}.".format(self.chars)
                     ctx.set(old, msg)
                     raise Exception(msg)
             return pos, (self.chars if self.value is self._NULL else self.value)
@@ -511,7 +568,7 @@ class Literal(Parser):
                     result.append(data[pos])
                     pos += 1
                 else:
-                    msg = "Expected case insensitive {!r}.".format(self.chars)
+                    msg = "Expected case insensitive {0!r}.".format(self.chars)
                     ctx.set(old, msg)
                     raise Exception(msg)
             return pos, ("".join(result) if self.value is self._NULL else self.value)
@@ -537,10 +594,12 @@ class Mark(object):
     in the input. Marks can give more context to a value transformed by mapped
     functions.
     """
-    def __init__(self, lineno, col, value):
+    def __init__(self, lineno, col, value, start, end):
         self.lineno = lineno
         self.col = col
         self.value = value
+        self.start = start
+        self.end = end
 
 
 class PosMarker(Wrapper):
@@ -552,8 +611,8 @@ class PosMarker(Wrapper):
     def process(self, pos, data, ctx):
         lineno = ctx.line(pos) + 1
         col = ctx.col(pos) + 1
-        pos, result = super(PosMarker, self).process(pos, data, ctx)
-        return pos, Mark(lineno, col, result)
+        newpos, result = super(PosMarker, self).process(pos, data, ctx)
+        return newpos, Mark(lineno, col, result, pos, newpos)
 
 
 class Sequence(Parser):
@@ -676,10 +735,11 @@ class Many(Parser):
             bs = Many(Char("b"), lower=1) # requires at least one "b"
 
     """
-    def __init__(self, parser, lower=0):
+    def __init__(self, parser, lower=0, upper=None):
         super(Many, self).__init__()
         self.add_child(parser)
         self.lower = lower
+        self.upper = upper
 
     def process(self, pos, data, ctx):
         orig = pos
@@ -693,7 +753,12 @@ class Many(Parser):
                 break
         if len(results) < self.lower:
             child = self.children[0]
-            msg = "Expected at least {} of {}.".format(self.lower, child)
+            msg = "Expected at least {0} of {1}.".format(self.lower, child)
+            ctx.set(orig, msg)
+            raise Exception()
+
+        if self.upper is not None and len(results) > self.upper:
+            msg = "Expected at most {0} of {1}.".format(self.upper, child)
             ctx.set(orig, msg)
             raise Exception()
 
@@ -701,7 +766,7 @@ class Many(Parser):
 
     def __repr__(self):
         if not self.name:
-            return "Many({}, lower={})".format(self.children[0], self.lower)
+            return "Many({0}, lower={1})".format(self.children[0], self.lower)
         return super(Many, self).__repr__()
 
 
@@ -728,11 +793,13 @@ class Until(Parser):
             val = cs("abcdycc")           # returns ["a", "b", "c", "d"]
 
     """
-    def __init__(self, parser, predicate):
+    def __init__(self, parser, predicate, upper=None):
         super(Until, self).__init__()
         self.set_children([parser, predicate])
+        self.upper = upper
 
     def process(self, pos, data, ctx):
+        bound = self.upper
         parser, pred = self.children
         results = []
         while True:
@@ -744,6 +811,10 @@ class Until(Parser):
                     results.append(res)
                 except Exception:
                     break
+                if bound is not None and len(results) > bound:
+                    msg = "{0} matched more than {1}.".format(self.parser, bound)
+                    ctx.set(pos, msg)
+                    raise Exception()
             else:
                 break
         return pos, results
@@ -805,7 +876,7 @@ class NotFollowedBy(Parser):
         except Exception:
             return new, res
         else:
-            msg = "{} can't follow {}".format(right, left)
+            msg = "{0} can't follow {1}".format(right, left)
             ctx.set(new, msg)
             raise Exception()
 
@@ -920,7 +991,7 @@ class Map(Parser):
 
     def __repr__(self):
         if not self.name:
-            return "Map({}({}))".format(self.func.__name__, self.children[0])
+            return "Map({0}({1}))".format(self.func.__name__, self.children[0])
         return super(Map, self).__repr__()
 
 
@@ -1159,7 +1230,7 @@ class EndTagName(Wrapper):
             e = expect.lower()
 
         if r != e:
-            msg = "Expected {!r}. Got {!r}.".format(expect, res)
+            msg = "Expected {0!r}. Got {1!r}.".format(expect, res)
             ctx.set(pos, msg)
             raise Exception(msg)
         return pos, res
@@ -1174,6 +1245,14 @@ def skip_none(x):
     return [i for i in x if i is not None]
 
 
+class Space(Parser):
+    def process(self, pos, data, ctx):
+        if data[pos].isspace():
+            return pos + 1, data[pos]
+        ctx.set(pos, "Expected whitespace character.")
+        raise Exception()
+
+
 EOF = EOF() % "EOF"
 EOL = InSet("\n\r") % "EOL"
 LineEnd = Wrapper(EOL | EOF) % "LineEnd"
@@ -1196,8 +1275,9 @@ Digit = InSet(string.digits) % "digit"
 Digits = String(string.digits) % "digits"
 Letter = InSet(string.ascii_letters) % "ASCII letter"
 Letters = String(string.ascii_letters) % "ASCII letters"
+Space = Space() % "any whitespace"
 WSChar = InSet(set(string.whitespace) - set("\n\r")) % "whitespace w/o EOL"
-WS = Many(InSet(string.whitespace) % "any whitespace")
+WS = Many(Space)
 Number = (Lift(_make_number) * Opt(Char("-"), "") * Digits * Opt(Char(".") + Digits)) % "number"
 SingleQuotedString = Char("'") >> String(set(string.printable) - set("'"), "'") << Char("'")
 DoubleQuotedString = Char('"') >> String(set(string.printable) - set('"'), '"') << Char('"')
